#!/usr/bin/env python
# -*- coding: utf-8 -*-

from builtins import str, zip, range
import os
import sys

import numpy as np
import pandas as pd
import pubchempy as pcp
import time
import re

from cornaconvertor import constants as cs
from datum import helpers as dat_hlp


dir_path = os.path.dirname(os.path.realpath(__file__))
dat_tim = time.strftime(cs.DATE_TIME)


def check_file_exist(path):
    """
    This function validates the path of file and raises
    error upon failure.
    :param path: path to input file.
    :return: Raises error upon failure.
    """
    if os.path.exists(path) == False:
        raise IOError(cs.ERR_FILE_DOESNOT_EXIST_MSG)


def check_empty_dataframe(input_df):
    """
    This function validates that the dataframe is not
    empty and raises an error upon failure.
    :param input_df: input Dataframe
    :return: raises error upon failure
    """
    if input_df.empty:
        raise IOError(cs.ERR_DF_EMPTY_MSG)


def get_readfile_func(path):
    """
    This functions fetches the extension of the file to
    be read and maps it to the corresponding read function
    in pandas.
    :param: path of the input file
    :return: Read function of pandas according to
             file extension.
    """
    file_ext = os.path.splitext(path)[1]
    read_func = cs.EXT_DICT.get(file_ext)
    if read_func == None:
        raise IOError(cs.ERR_FILE_EXTN_MSG)
    return read_func


def read_file(path):
    """
    This function reads the input file in xls, xlsx, txt and csv
    format and returns a dataframe.
    Param:path : path to input file
    Returns:input_file_df : input file in the form of pandas dataframe
    """
    check_file_exist(path)
    if os.stat(path).st_size == 0:
        raise IOError(cs.ERR_EMPTY_FILE_MSG)
    else:
        read_function = get_readfile_func(path)
        input_file_df = read_function(path, header=0, index_col=False)
    check_empty_dataframe(input_file_df)
    return input_file_df


def check_no_formula_column_exist(input_df):
    """
    This function checks if the Formula column
    exists in the dataframe.
    Param: input_df: Input dataframe
    Return: Boolean (True or False)
    """

    return cs.FORMULA_COLUMN not in input_df.columns

def check_no_label_column_exist(input_df):
    """
    This function checks if the label column
    exists in the dataframe.
    Param: input_df: Input dataframe
    Return: Boolean (True or False)
    """
    return cs.SMALL_L_LABEL_COLUMN not in input_df.columns


def fetch_chemical_formula_web(compound_name):
    try:
        for compound in pcp.get_compounds(compound_name, 'name'):
            c = pcp.Compound.from_cid(compound.cid)
            return c.molecular_formula
    except ValueError:
        return None


def fetch_chemical_formula(compound_name):
    """
    Fetch formula of the given compound name from the formula file, if not present in the
    formula file then conduct a web search.
    :param compound_name: name of the compound whose formula is to be fetched
    :return: formula of the compound
    """
    path = os.path.join(dir_path, cs.CHEMICAL_FORMULA_FILE)
    formula_df = pd.read_csv(path)
    compound_details = formula_df[formula_df.Name == compound_name]\
        [cs.CHEMICAL_FORMULA_COLUMN].get_values()

    try:
        chemical_formula = compound_details[0]
        return chemical_formula
    except IndexError:
        return fetch_chemical_formula_web(compound_name)


def check_columns(input_df):
    """
    This function validates the presence of required columns in the input file.
    It raises error upon failure.
    :param input_df: Dataframe of the input file
    :return: raises error if required columns are not present.
    """
    required_cols_set = set([x.lower() for x in cs.REQUIRED_COL_LIST])
    alternate_required_cols_set = set([x.lower() for x in cs.NEW_REQUIRED_COL_LIST])
    set_df = set([x.lower() for x in list(input_df)])
    if required_cols_set.issubset(set_df) or\
        alternate_required_cols_set.issubset(set_df):
        pass
    else:
        raise IOError(cs.ERR_MISSING_COL_MSG)


def check_column_headers(col_headers,col_names):
    """
    This function verifies that all default columns are present in input text files
    Args :
        col_headers : Column headers which are present in the dataframe
        col_names : Column headers for which we need to check if they are present or not
    return:
        error message if col_names is not subset of col_headers, else do nothing
    """
    err_msg = """Required column/s not found, Column: {!r}""". \
        format(list(set(col_names) - set(col_headers)))
    assert set(col_names).issubset(set(col_headers)),err_msg


def remove_cit(df,col_name):
    """
    This function replaces all the 'G9 0min cit', with 'G9 0min' in column 'Sample'. it is
    specifically of El-MAVEN output of MIMOSA experiment
    Args:
        df : dataframe
        col_name : name of column where it would search for the string
    return:
        updated column with 'cit' removed
    """
    df[col_name] = df[col_name].str.replace(' cit','')
    return df[col_name]


def write_to_csv(input_file_path,file_name,df):
    """
    This function will be used to generate an csv file
    Args:
        dir: current working directory
        file : name of output csv file
        df : dataframe which would be written in the csv file
    return:
        csv file after writing dataframe
    """
    dir = os.path.dirname(input_file_path)
    na_in = os.path.join(dir, file_name)
    df.to_csv(na_in,sep=',',header=True,index=False)
    return na_in


def rename_columns(df,old_name,new_name):
    """
    This function will rename column names
    Args:
        df : dataframe on which col names needs to be changed
        old_name : Old column names
        new_name : New column names
    return:
        Changes the requested col names and returns the updated dataframe
    """
    if len(old_name) == 0:
        df = df.rename(columns={old_name: new_name},inplace=True)
    else:
        df = df.rename(columns=dict(list(zip(old_name,new_name))),inplace=True)
    return df


def get_uniques_values(df,col_name):
    """
    This function will extract all the unique values from the column of a dataframe
    Args:
        df : Dataframe
        col_name : Column name from where unique values will be extracted
    return:
        list of all the unique values in that column
    """
    unique_val = np.unique(df[[col_name]].values)
    return unique_val


def filter_col(df,col_name,text):
    """
    This function will filter the column in a dataframe on the basis of text provided by user
    Args:
        df: dataframe
        col_name: Column name where it would apply filter
        text: text with which it would apply the filter
    return:
        filtered column
    """
    new_df = df.loc[df[col_name] == text]
    return new_df


def get_mass_info(dir):
    """Creates 'Mass Info' column which is mandatory for NA Correction from MIMOSA Compound
    Database stored in data_files
    csants:
    MASS_INFO1 = 'precursorMz'
    MASS_INFO2 = 'productMz'
    MASS_INFO = 'Mass Info'
    Args:
    dir : Current working directory
    return :
    It returns a dataframe containing Mass Info of all the components
    """
    path = os.path.join(dir, cs.COMP_DATA)
    comp_data_df = pd.read_csv(path, header=0, index_col=False)
    rename_columns(comp_data_df, [cs.MVN_COL[0]], [cs.NA_COL[0]])
    comp_data_df[cs.MASS_INFO1] = comp_data_df[cs.MASS_INFO1].astype('float64')
    comp_data_df[cs.MASS_INFO2] = comp_data_df[cs.MASS_INFO2].astype('float64')
    comp_data_df[cs.MASS_INFO] = comp_data_df.apply(
        lambda x: '%s / %s' % (x[cs.MASS_INFO1], x[cs.MASS_INFO2]), axis=1)
    mass_info = comp_data_df[[cs.NA_COL[0], cs.MASS_INFO]]
    mass_info.drop_duplicates(subset=cs.NA_COL[0], keep='first', inplace=True)
    return mass_info


def get_mass_info_for_RNA_data(dir):
    """Creates 'Mass Info' column which is mandatory for NA Correction from MIMOSA Compound
    Database stored in data_files
    csants:
    MASS_INFO1 = 'precursorMz'
    MASS_INFO2 = 'productMz'
    MASS_INFO = 'Mass Info'
    Args:
    dir : Current working directory
    return :
    It returns a dataframe containing Mass Info of all the components
    """
    path = os.path.join(dir)
    comp_data_df = pd.read_csv(path, header=0, index_col=False)
    rename_columns(comp_data_df, [cs.MVN_COL[0]], [cs.NA_COL[0]])
    comp_data_df[cs.MASS_INFO1] = comp_data_df[cs.MASS_INFO1].astype('float64')
    comp_data_df[cs.MASS_INFO2] = comp_data_df[cs.MASS_INFO2].astype('float64')
    comp_data_df[cs.MASS_INFO] = comp_data_df.apply(
        lambda x: '%s / %s' % (x[cs.MASS_INFO1], x[cs.MASS_INFO2]), axis=1)
    mass_info = comp_data_df[[cs.NA_COL[0], cs.MASS_INFO]]
    return mass_info


def get_method_type(dir):
    """
    This function would read "metadata_mq_with_method_all.csv" and then create a dataframe for
    method type ofl different metabolites and then merges it with mass_info dataframe creating
    mass_info_with_method_df
    Args:
        dir: Current working directory
    return:
        mass_info_with_method_df dataframe
    """
    path = os.path.join(dir, cs.METADATA_FILE)
    metadata_df = pd.read_excel(path, header=0)
    method_df = metadata_df[cs.METADATA_COL]
    method_df[cs.METADATA_COL[1]] = method_df[cs.METADATA_COL[1]].astype(str)
    return method_df


# def add_missing_comp(df, method):
#     """
#     This function would search for missing 'Component Name' in each filename present in 'Original
#     Filename' column, and then add additional rows containing required information for the missing
#     'Component Name' in the dataframe.
#     'Component Name' : Column name in the dataframe, containing different components of a fragment
#     csants :
#     NA_COL = ['Component Name', 'Original Filename', 'Area']
#     Args:
#         df : dataframe on which this function would be applied
#         method : 'Method type' which would be entered in the newly added row
#     Return:
#         New dataframe after adding missing components
#     """
#     df.reset_index(inplace = True, drop = True)
#     uniq_comp = get_uniques_values(df, cs.NA_COL[0])
#     uniq_file = get_uniques_values(df, cs.NA_COL[1])
#     col_names = cs.NA_COL + [cs.METADATA_COL[1]]
#     df_new = pd.DataFrame(data=None,columns=col_names)
#     for file in uniq_file:
#         df_file = filter_col(df, cs.NA_COL[1], file)
#         uniq_comp_in_file = get_uniques_values(df_file, cs.NA_COL[0])
#         if set(uniq_comp) - set(uniq_comp_in_file) == set([]):
#             pass
#         else:
#             miss_comp = list(set(uniq_comp) - set(uniq_comp_in_file))
#             for i in range(len(miss_comp)):
#                 df.loc[len(df.index)] = [miss_comp[i], file, 0.00, method]

#     df_new = pd.concat([df_new, df])
#     return df_new


def add_missing_comp(df, method):
    """
    This function would search for missing 'Component Name' in each filename present in 'Original
    Filename' column, and then add additional rows containing required information for the missing
    'Component Name' in the dataframe.
    'Component Name' : Column name in the dataframe, containing different components of a fragment
    csants :
    NA_COL = ['Component Name', 'Original Filename', 'Area']
    Args:
        df : dataframe on which this function would be applied
        method : 'Method type' which would be entered in the newly added row
    Return:
        New dataframe after adding missing components
    """
    df.reset_index(inplace=True, drop=True)
    uniq_file = get_uniques_values(df, cs.NA_COL[1])
    df = get_isotopomers_of_metabolite(df, method, uniq_file[1] )
    uniq_comp = get_uniques_values(df, cs.NA_COL[0])
    col_names = cs.NA_COL + [cs.METADATA_COL[1]]
    df_new = pd.DataFrame(data=None,columns=col_names)
    for file in uniq_file:
        df_file = filter_col(df, cs.NA_COL[1], file)
        uniq_comp_in_file = get_uniques_values(df_file, cs.NA_COL[0])
        if set(uniq_comp) - set(uniq_comp_in_file) == set([]):
            pass
        else:
            miss_comp = list(set(uniq_comp) - set(uniq_comp_in_file))
            for comp_name in miss_comp:
                df.loc[len(df.index)] = [comp_name, file, 0.00, method]
    df_new = pd.concat([df_new,df])
    return df_new

def add_missing_comp_for_RNA_data(df, method, metadata_file_path):
    """
    This function would search for missing 'Component Name' in each filename present in 'Original
    Filename' column, and then add additional rows containing required information for the missing
    'Component Name' in the dataframe.
    'Component Name' : Column name in the dataframe, containing different components of a fragment
    csants :
    NA_COL = ['Component Name', 'Original Filename', 'Area']
    Args:
        df : dataframe on which this function would be applied
        method : 'Method type' which would be entered in the newly added row
    Return:
        New dataframe after adding missing components
    """
    df.reset_index(inplace=True, drop=True)
    uniq_file = get_uniques_values(df, cs.NA_COL[1])
    df = get_isotopomers_of_metabolite_for_RNA_data(df, method, uniq_file[1], metadata_file_path)
    uniq_comp = get_uniques_values(df, cs.NA_COL[0])
    col_names = cs.NA_COL + [cs.METADATA_COL[1]]
    df_new = pd.DataFrame(data=None,columns=col_names)
    for file in uniq_file:
        df_file = filter_col(df, cs.NA_COL[1], file)
        uniq_comp_in_file = get_uniques_values(df_file, cs.NA_COL[0])
        if set(uniq_comp) - set(uniq_comp_in_file) == set([]):
            pass
        else:
            miss_comp = list(set(uniq_comp) - set(uniq_comp_in_file))
            for i in range(len(miss_comp)):
                df.loc[len(df.index)] = [miss_comp[i],file,0.00,method]
    df_new = pd.concat([df_new,df])
    return df_new


def get_isotopomers_of_metabolite_for_RNA_data(df, method, sample_name, metadata_file_path):
    """
    This function would check if any particular isotopomer of a fragment is missing  the input
    files, and after finding the missing isotopomer, it adds a row containing all the relevent info-
    -rmation about that isotopomer in the df, taking intensity as zero
    :param dir:(type: string) path of package directory
    :param df:(type: pd.Dataframe) Dataframe generated from input file
    :param method:(type: string) method type for experiment
    :param sample_name:(type: string) Name of sample
    :return:df(type: pd.Dataframe) Updated dataframe
    """
    path = os.path.join(metadata_file_path)
    metadata_df = pd.read_excel( path,header=0 )
    df.reset_index( inplace=True,drop=True )
    uniq_meta_frag = get_uniques_values( metadata_df,cs.METADATA_COL[2])
    frag_dict = {}
    for frag in uniq_meta_frag:
        list_isotopomer = list(metadata_df.loc[metadata_df[cs.METADATA_COL[2]] == frag]
                               [cs.METADATA_COL[0]].unique())
        frag_dict[frag] = list_isotopomer
    df[cs.COMP_AREA[2]] = df[cs.METADATA_COL[0]].apply(lambda x:search_dict(frag_dict,x))
    df = df[~df.astype(str).eq('None').any(1)]
    uniq_input_frag = get_uniques_values(df, cs.COMP_AREA[2])


    input_frag_dict = {}
    for frag in uniq_input_frag:
        list_isotopomer = list(df.loc[df[cs.COMP_AREA[2]] == frag][cs.METADATA_COL[0]].
                               unique())
        input_frag_dict[frag] = list_isotopomer
    for frag in uniq_input_frag:
        if set(frag_dict[frag]) - set(input_frag_dict[frag]) == set([]):
            print("All isotopomers of fragment {} are present".format(frag))
        else:
            miss_comp = list(set(frag_dict[frag]) - set(input_frag_dict[frag]))
            print("These isotopomers of {} are missing: {!r}".format(frag, miss_comp))
            for i in range(len(miss_comp)):
                df.loc[len(df.index)] = [miss_comp[i],sample_name,0.00,method, frag]
    df = df.drop(cs.COMP_AREA[2],1)
    return df

def get_isotopomers_of_metabolite(df, method, sample_name):
    """
    This function would check if any particular isotopomer of a fragment is missing  the input
    files, and after finding the missing isotopomer, it adds a row containing all the relevent info-
    -rmation about that isotopomer in the df, taking intensity as zero
    :param dir:(type: string) path of package directory
    :param df:(type: pd.Dataframe) Dataframe generated from input file
    :param method:(type: string) method type for experiment
    :param sample_name:(type: string) Name of sample
    :return:df(type: pd.Dataframe) Updated dataframe
    """

    df.reset_index( inplace=True,drop=True )

    #read metadata file and create a copy with stripped frag names
    path = os.path.join(dir_path, cs.METADATA_FILE)
    metadata_df = pd.read_csv( path,header=0 )
    cloned_metadata = metadata_df.copy()
    cloned_metadata[cs.METADATA_COL[2]] =  cloned_metadata[cs.METADATA_COL[2]].apply(clean_string)
    cloned_metadata[cs.METADATA_COL[0]] = cloned_metadata[cs.METADATA_COL[0]].apply(clean_string)

    #create a dict of parent frag and list of related isotopomer fragments
    uniq_meta_frag = get_uniques_values(cloned_metadata, cs.METADATA_COL[2])
    meta_frag_dict = {}
    for frag in uniq_meta_frag:
        list_isotopomer = list(cloned_metadata.loc[cloned_metadata[cs.METADATA_COL[2]] == frag]
                              [cs.METADATA_COL[0]].unique())
        meta_frag_dict[frag] = list_isotopomer

    #add column for parent frag in input df and create a copy
    df[cs.COMP_AREA[2]] = df[cs.METADATA_COL[0]].apply(lambda x:search_dict(meta_frag_dict,x))
    df = df[~df.astype(str).eq('None').any(1)]
    cloned_input = df.copy()
    df = df.drop(cs.COMP_AREA[2],1)
    cloned_input[cs.METADATA_COL[0]] = cloned_input[cs.METADATA_COL[0]].apply(clean_string)
    cloned_input[cs.COMP_AREA[2]] = cloned_input[cs.COMP_AREA[2]].apply(clean_string)

    #create a dict of parent frag and list of related isotopomer fragments in input
    uniq_input_frag = get_uniques_values(cloned_input, cs.COMP_AREA[2])
    input_frag_dict = {}
    for frag in uniq_input_frag:
        list_isotopomer = list(cloned_input.loc[cloned_input[cs.COMP_AREA[2]] == frag][cs.METADATA_COL[0]].
                               unique())
        input_frag_dict[frag] = list_isotopomer

    #find and add entries for missing fragments in input df
    for frag in uniq_input_frag:
        if set(meta_frag_dict[frag]) - set(input_frag_dict[frag]) == set([]):
            print("All isotopomers of fragment {} are present".format(frag))
        else:
            miss_comp = list(set(meta_frag_dict[frag]) - set(input_frag_dict[frag]))
            print("These isotopomers of {} are missing: {!r}".format(frag, miss_comp))
            for comp in miss_comp:
                name = list(metadata_df.loc[cloned_metadata[cs.METADATA_COL[0]] == comp]
                            [cs.METADATA_COL[0]])[0]
                df.loc[len(df.index)] = [name, sample_name, 0.00, method]
    return df


def search_dict(myDict,lookup_val):
    """
    This function would search for a particular string inside a dictionary
    :param myDict:(type: dict) dictionary in which the function will search for string
    :param lookup_val:(type: string) lookup value
    :return: key(type: string) keys which would contain the lookup value
    """
    for key, value in list(myDict.items()):
        found = [v for v in value if clean_string(lookup_val) == clean_string(v)]
        if found:
            return key

def remove_column_from_df(df, column_name):
    df = df.drop(column_name, axis=1)
    return df

def clean_string(text):
    """
    This function removes all whitespaces from the given string and converts it
    to lower case
    :param text:(type: string) string to be cleaned
    :return (type: string) lower case, cleaned string without any whitespace
    """
    return re.sub(r'\s+', '', text).lower()

def replace_kegg_with_hmdb(df, kegg_list):
    """
    This function checks for kegg id in dataframe and returns its hmdb id if
    present in dataframe

    :param df: (Dataframe) Having information of KEGG, HMDB and Metabolite name
    :param kegg_list: (list) list of keggID
    :return: Returns list of hmdb corresponding to kegg
    """
    hmdb = []
    for item in kegg_list:
        temp = ""
        for index, rows in df.iterrows():
            if item == rows[cs.KEGG_COLUMN]:
                temp = rows[cs.HMDBID_COLUMN]
        if temp == "":
            hmdb.append(item)
        else:
            hmdb.append(temp)
    return hmdb


def convert_primary_to_sec_hmdb(primary_id):
    """
    This function converts primary hmdb ID's of length 11 to secondary hmdb ID's
    of length 9.

    :param primary_id: pandas Series
        compound ID column of pandas dataframe as pandas series

    :return:
        converted pandas series
    """
    primary_id = str(primary_id)
    sec_id = primary_id[:4] + primary_id[-5:]
    return sec_id


def get_sec_hmdb_df(good_peak_df):
    """
    This function checks length of hmdb ID, if length 9 then it is
    retained in the dataframe and if length 11 then converted to length 9 (
    secondary hmdb ID) in the dataframe.

    :param good_peak_df: pandas DataFrame
        Maven file dataframe having good peaks

    :return:
    pandas dataframe with secondary hmdb ID
    """
    good_df = good_peak_df[(good_peak_df[cs.COMPOUNDID_COLUMN].str.len() == 11)]
    good_df[cs.COMPOUNDID_COLUMN] = good_df[cs.COMPOUNDID_COLUMN].apply(
        convert_primary_to_sec_hmdb)
    sec_index = good_peak_df.index.difference(good_df.index)
    left_df = good_peak_df.ix[sec_index, :]
    sec_hmdb_df = pd.concat([good_df, left_df], ignore_index=True)
    return sec_hmdb_df

def match_metabolites_from_db_for_msms(input_file_path):
    """
    This function matches with the metabolites present on the input dataframe and
    the database. If any metabolite does not match with the database, it reports the 
    missing metabolites as a list in error.log file.
    Which will be popped out to user in front-end.
    
    :param input_df pandas dataframe
        Input file from user.
    
    :return:
        Throws a list of missing metbolites in error.log
    """
    input_df = dat_hlp.read_file(input_file_path)
    path = os.path.join(dir_path, cs.COMP_DATA)
    mimosa_db = dat_hlp.read_file(path, header=0, index_col=False)
    non_matching_metabolites = []
    non_matching_metabolites = list(set(input_df[cs.COMPOUND_COLUMN])-set(mimosa_db[cs.COMPOUND_COLUMN]))
    return non_matching_metabolites