#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import pandas as pd
import time
import numpy as np
from cornaconvertor import constants as const
from cornaconvertor import helpers as hlp
from datum import helpers as dat_hlp

pd.options.mode.chained_assignment = None
dat_tim = time.strftime(const.DATE_TIME)
dir_path = os.path.dirname(os.path.realpath(__file__))


def convert_file_LCMS(input_file_path):
    """
    This function reads input file and convert it into a dataframe.
    Modify dataframe according to NA correction tool :  Rename columns,
    add formula column, drop columns.
    Write dataframe into a csv file.
    :param input_file_path: path to input file.
    :return: list of path of .csv file compatible to NA correction tool.
    """
    input_df = dat_hlp.read_file(input_file_path)
    input_df = input_df.dropna(how='all')
    if hlp.check_no_label_column_exist(input_df):
        input_df[const.SMALL_L_LABEL_COLUMN] = None
    
    else :
        try:
            bad_groups = list(input_df[(input_df.label == 'b')&(input_df.isotopeLabel == 'C12 PARENT')]['metaGroupId'])
            if(bad_groups):
                input_df = input_df[~input_df['metaGroupId'].isin(bad_groups)]   
            input_df = input_df[input_df[const.SMALL_L_LABEL_COLUMN] != "b"]
        except:
            pass

    hlp.check_columns(input_df)
    input_df = hlp.remove_column_from_df(input_df, const.SMALL_L_LABEL_COLUMN)
    input_df = input_df.rename(
        columns={const.COMPOUND_COLUMN: const.NAME_COLUMN,
                 const.NOTE_COLUMN:const.LABEL_COLUMN,
                 const.ISOTOPELABEL_COLUMN:const.LABEL_COLUMN
                 })

    if hlp.check_no_formula_column_exist(input_df):
        input_df[const.FORMULA_COLUMN] = input_df[const.NAME_COLUMN].apply(hlp.fetch_chemical_formula)
        input_df.rename(columns = {const.FORMULA_COLUMN : 'Formula'}, inplace=True)
    else:
        input_df.rename(columns = {const.FORMULA_COLUMN : 'Formula'}, inplace=True)

    input_df.drop(const.EXTRA_COLUMNS_LIST, axis=1, inplace=True)
    input_df.dropna()
    return input_df


def convert_file_MSMS(input_file_path,metadata_file_path):
    """
    This function takes the path of the input file that is to be converted.
    it takes raw_df as input,and then after checking for the required col names
    performs various operations required to convert data to be NA Correction-ready,
    i.e., renaming columns, removing strings, merging
    dataframes, thus resulting into a final dataframe which is then broken in sub-dataframes on the
    basis of 'Method-Type', these dataframes are then further exported to .csv file
    :param 
        input_file_path: path to input file.
        metadata_file_path: path to the user metadata file.
    :return: list of paths of .csv file compatible to NA correction tool.
    """
    logs = {const.ERRORS: [], const.WARNINGS: {const.MESSAGE: [], const.ACTION: []}}
    input_df = dat_hlp.read_file(input_file_path)
    user_metadata = dat_hlp.read_file(metadata_file_path)

    internal_metadata_db_path = os.path.join(dir_path, const.METADATA_FILE)
    internal_metadata = dat_hlp.read_file(internal_metadata_db_path)
    
    mimosa_db_path = os.path.join(dir_path, const.COMP_DATA)
    mimosa_db = dat_hlp.read_file(mimosa_db_path)
    
    #updated metadata
    new_fragments_metadata = pd.DataFrame()
    new_fragments_metadata = internal_metadata[~internal_metadata[const.COMPONENT_NAME_COLUMN].isin(user_metadata[const.COMPONENT_NAME_COLUMN])].dropna()
    user_metadata[const.METHOD] = const.CIT
    internal_metadata = new_fragments_metadata.append(user_metadata)
    internal_metadata.to_csv(internal_metadata_db_path,index =False)
    
    #updated mimosa_db
    if not user_metadata.empty:
        add_mimosa_fragments = pd.DataFrame()
        fragments = user_metadata[const.COMPONENT_NAME_COLUMN].str.split()
        precursor_product_mz = fragments.str[-1].str.split('/', expand=True)
        add_mimosa_fragments[const.COMPOUND_COLUMN] = user_metadata[const.COMPONENT_NAME_COLUMN]
        add_mimosa_fragments[const.MASS_INFO1] = precursor_product_mz[0]
        add_mimosa_fragments[const.MASS_INFO2] = precursor_product_mz[1]
        mimosa_db = mimosa_db.append(add_mimosa_fragments)
        mimosa_db.drop_duplicates(subset =const.COMPOUND_COLUMN, inplace = True)
        mimosa_db.to_csv(mimosa_db_path,index = False)

    col_headers = input_df.columns.tolist()
    hlp.check_column_headers(col_headers, const.MVN_COL)
    hlp.rename_columns(input_df, const.MVN_COL, const.NA_COL)
    input_df[[const.ORIGINAL_FILENAME]] = input_df[[const.ORIGINAL_FILENAME]].astype(str)
    try:
        input_df[const.NA_COL[1]] = hlp.remove_cit(input_df, const.NA_COL[1])
    except Exception as e:
        print(str(e))
    mvn_df = input_df[const.NA_COL]
    # TODO: Commenting logic to remove std samples - NCT-623
    # mvn_df = mvn_df[~mvn_df[const.NA_COL[1]].str.contains("std")]
    # if mvn_df.duplicated(subset=const.NA_COL[0:2]).any():
    #     logs[const.WARNINGS][const.MESSAGE] = "Warning : The input file", file_name,\
    #         "contains duplicates in columns :",\
    #         const.NA_COL[0:2], ", which would be removed"
    mvn_df = mvn_df.drop_duplicates(subset=const.NA_COL[0:2], keep= 'first')
    mvn_df[const.METADATA_COL[1]] = 'cit'

    df_method = hlp.filter_col(mvn_df, const.METADATA_COL[1], const.CIT)
    df_method = hlp.add_missing_comp(df_method, const.CIT)
    df_mass_info = hlp.get_mass_info(dir_path)
    df_mass_info["temp"] = df_mass_info[const.NA_COL[0]].apply(hlp.clean_string)
    df_mass_info.drop(const.NA_COL[0], axis = 1, inplace = True)
    df_method["temp"] = df_method[const.NA_COL[0]].apply(hlp.clean_string)
    df_final = pd.merge(df_method, df_mass_info, on="temp")
    output_file_name = const.OUTPUT_FILE + const.CSV_EXT
    csv_output = hlp.write_to_csv(input_file_path, output_file_name, df_final)
    df_final.drop(["temp", const.METADATA_COL[1]], axis = 1, inplace = True)
    df_final['Sample Name'] = df_final[[const.NA_COL[1]]]
    return df_final


def convert_msms_quant(input_file_path):
    """
    This function converts LCMS/MS maven output to quant compatible
    input
    :param input_file_path: path of the file to be converted
    :return: quant compatible dataframe
    """
    input_df = dat_hlp.read_file(input_file_path)

    if const.FORMULA_COLUMN in input_df:
        input_df = hlp.remove_column_from_df(input_df, const.FORMULA_COLUMN)
    elif const.UPPERCASE_FORMULA_COLUMN in input_df:
        input_df = hlp.remove_column_from_df(input_df, const.UPPERCASE_FORMULA_COLUMN)

    col_headers = input_df.columns.tolist()
    hlp.check_column_headers(col_headers, const.MVN_COL)
    hlp.rename_columns(input_df, const.MVN_COL, const.NA_COL)
    input_df[const.NA_COL[1]] = hlp.remove_cit(input_df, const.NA_COL[1])
    mvn_df = input_df[const.NA_COL]
    df_mass_info = hlp.get_mass_info(dir_path)
    mvn_df_1 = pd.merge(mvn_df, df_mass_info, on=const.NA_COL[0])
    mvn_df_1[const.SAMPLE_NAME] = mvn_df_1[const.NA_COL[1]]
    return mvn_df_1


def convert_lcms_quant(input_file_path):
    """
    This function converts LCMS maven output to quant compatible input.
    :param input_file_path: path of the file to be converted
    :return: quant compatible dataframe
    """
    input_df = dat_hlp.read_file(input_file_path)

    if const.FORMULA_COLUMN in input_df.columns:
        input_df = hlp.remove_column_from_df(input_df, const.FORMULA_COLUMN)
    elif const.UPPERCASE_FORMULA_COLUMN in input_df.columns:
        input_df = hlp.remove_column_from_df(input_df, const.UPPERCASE_FORMULA_COLUMN)

    hlp.check_columns(input_df)
    input_df = input_df.rename(columns={const.COMPOUND_COLUMN: const.NA_COL[0]})
    input_df.drop(const.EXTRA_COLUMNS_LIST_QUANT, axis=1, inplace=True, errors = 'ignore')
    input_df.dropna()
    input_df = pd.melt(input_df, id_vars=[const.NA_COL[0]], var_name=const.NA_COL[1],
                       value_name=const.NA_COL[2])
    input_df[const.SAMPLE_NAME] = input_df[const.NA_COL[1]]
    input_df[const.MASS_INFO] = 0
    return input_df


def convert_maven_to_sample_qc(input_file_path):
    """
    This function converts the maven output file into a
    file format compatible with sample qc input requirement.
    :param input_file_path: path of the file to be converted
    :return: sample_qc compatible dataframe
    """
    input_df = dat_hlp.read_file(input_file_path)
    hlp.check_columns(input_df)
    if 'formula' in input_df:
        input_df.drop(const.EXTRA_COLUMNS_LIST_SAMPLE_QC, axis=1, inplace=True)
    else:
        input_df.drop(const.EXTRA_COLUMNS_LIST_SAMPLE_QC_OLD, axis=1, inplace=True)
    input_df.dropna()
    if const.NOTE_COLUMN in input_df:
        input_df[const.SAMPLE_ID] = input_df[const.NOTE_COLUMN] + input_df[const.COMPOUND_COLUMN]
        input_df.drop([const.NOTE_COLUMN, const.COMPOUND_COLUMN], axis=1, inplace=True)
    else :
        input_df[const.SAMPLE_ID] = input_df[const.ISOTOPELABEL_COLUMN] + input_df[
            const.COMPOUND_COLUMN]
        input_df.drop([const.ISOTOPELABEL_COLUMN, const.COMPOUND_COLUMN], axis=1,
                      inplace=True)
    list_cols = input_df.columns.tolist()
    sample_col_first_list = list_cols[-1:] + list_cols[:-1]
    input_df = input_df[sample_col_first_list]
    input_df = input_df.set_index(const.SAMPLE_ID).transpose()
    input_df.index.name = const.SAMPLE_ID
    input_df = input_df.reset_index()

    return input_df

def convert_maven_to_unlabelled_pvd(input_file_path, keep_duplicate_name=False):
    """
    This function converts LCMS maven output to metscape compatible input.

    :param input_file_path: (string) path of the file to be converted
    :return: (dataframe) Metscape compatible df
    """
    input_df = dat_hlp.read_file(input_file_path)
    hlp.check_columns(input_df)
    df = input_df.dropna(how='all')
    df[const.SMALL_L_LABEL_COLUMN] = df[const.SMALL_L_LABEL_COLUMN].replace([np.nan," "], "g")
    df = df[df[const.SMALL_L_LABEL_COLUMN] == "g"]
    sample_df = df.dropna(subset=[const.COMPOUNDID_COLUMN], how='all')
    if const.NOTE_COLUMN in sample_df:
        if 'formula' in sample_df:
            if(keep_duplicate_name):
                sample_df = sample_df.drop(const.EXTRA_COLUMNS_LIST_UL_PVD_KEEP_DUPLICATE_NAMES, axis=1)
            else:
                sample_df = sample_df.drop(const.EXTRA_COLUMNS_LIST_UL_PVD, axis=1)
        else:
            if(keep_duplicate_name):
                sample_df = sample_df.drop(const.EXTRA_COLUMNS_LIST_UL_PVD_OLD_KEEP_DUPLICATE_NAMES, axis=1)
            else:
                sample_df = sample_df.drop(const.EXTRA_COLUMNS_LIST_UL_PVD_OLD, axis=1)
    else:
        if(keep_duplicate_name):
            sample_df = sample_df.drop(const.EXTRA_COLUMNS_LIST_UL_PVD_NEW_KEEP_DUPLICATE_NAMES, axis=1)
        else:
            sample_df = sample_df.drop(const.EXTRA_COLUMNS_LIST_UL_PVD_NEW, axis=1)
    if(not(keep_duplicate_name)):
        sample_df.set_index(const.COMPOUNDID_COLUMN, inplace=True)
        final_df = sample_df.transpose()
        final_df.index.name = const.SAMPLE_ID
        final_df = final_df.reset_index()
        return final_df
    else:
        return sample_df

def convert_RNA_file_MSMS(input_files_list):
    """
    This function takes the path of the input file that is to be converted.
    it takes raw_df as input,and then after checking for the required col names
    performs various operations required to convert data to be NA Correction-ready,
    i.e., renaming columns, removing strings, merging
    dataframes, thus resulting into a final dataframe which is then broken in sub-dataframes on the
    basis of 'Method-Type', these dataframes are then further exported to .csv file
    :param input_file_path: path to input file.
    :return: list of paths of .csv file compatible to NA correction tool.
    """
    input_file_path = input_files_list['raw_input_file']
    comp_file_path = input_files_list['comp_input_file']
    metadata_file_path = input_files_list['meta_input_file']
    logs = {const.ERRORS: [], const.WARNINGS: {const.MESSAGE: [], const.ACTION: []}}
    input_df = dat_hlp.read_file(input_file_path)
    col_headers = input_df.columns.tolist()
    hlp.check_column_headers(col_headers, const.MVN_COL)
    hlp.rename_columns(input_df, const.MVN_COL, const.NA_COL)
    input_df[[const.ORIGINAL_FILENAME]] = input_df[[const.ORIGINAL_FILENAME]].astype(str)
    try:
        input_df[const.NA_COL[1]] = hlp.remove_cit(input_df, const.NA_COL[1])
    except Exception as e:
        print(str(e))
    mvn_df = input_df[const.NA_COL]
    # TODO: Commenting logic to remove std samples - NCT-623
    # mvn_df = mvn_df[~mvn_df[const.NA_COL[1]].str.contains("std")]
    # if mvn_df.duplicated(subset=const.NA_COL[0:2]).any():
    #     logs[const.WARNINGS][const.MESSAGE] = "Warning : The input file", file_name,\
    #         "contains duplicates in columns :",\
    #         const.NA_COL[0:2], ", which would be removed"
    mvn_df = mvn_df.drop_duplicates(subset=const.NA_COL[0:2], keep= 'first')
    mvn_df[const.METADATA_COL[1]] = 'cit'
    df_method = hlp.filter_col(mvn_df, const.METADATA_COL[1], const.CIT)
    # df_method = hlp.add_missing_comp(df_method, const.CIT)
    df_method = hlp.add_missing_comp_for_RNA_data(df_method, const.CIT, metadata_file_path)

    df_mass_info = hlp.get_mass_info_for_RNA_data(comp_file_path)
    df_final = pd.merge(df_method, df_mass_info, on=const.NA_COL[0])
    output_file_name = const.OUTPUT_FILE + const.CSV_EXT
    csv_output = hlp.write_to_csv(input_file_path, output_file_name, df_final)
    df_final.drop(const.METADATA_COL[1], axis = 1, inplace = True)
    df_final['Sample Name'] = df_final[[const.NA_COL[1]]]
    # df_final.to_csv('~/Desktop/xxx2.csv', index=False)
    return df_final

def conversion(inputfile, filetype):
    """
    This function the file type to the function which i used for conversion.
    :param inputfile: file to be converted
    :param filetype: conversion type
    :return: converted data frame
    """

    function = const.CONVERSION_DICT.get(filetype)
    converted_df = globals()[function](inputfile)
    return converted_df

def convert_isocorrect_to_quant_lcms(input_file_path):
    """
    This function converts Isocorrect output to Quantfit compatible input.

    :param input_file_path: (string) path of the file to be converted
    :return: (dataframe) Quantfit compatible df
    """
    input_df = dat_hlp.read_file(input_file_path)
    output_df = pd.DataFrame()
    output_df[const.COMPONENT_NAME_COLUMN]= input_df[const.NAME_COLUMN]
    output_df[const.ORIGINAL_FILENAME]= input_df[const.SAMPLE_COLUMN]
    output_df[const.AREA_COLUMN]= input_df[const.NA_CORRECTED_WITH_ZERO_COLUMN]
    output_df[const.SAMPLE_NAME]= input_df[const.SAMPLE_COLUMN]
    output_df[const.MASS_INFO]= 0
    return output_df


def convert_isocorrect_to_quickcheck_lcms(input_file_path):
    """
    This function converts Isocorrect output to Quickcheck compatible input.

    :param input_file_path: (string) path of the file to be converted
    :return: (dataframe) Quickcheck compatible df
    """
    input_df = dat_hlp.read_file(input_file_path)
    input_df.drop(const.EXTRA_COLUMNS_LIST_ISOCORRECT_OUTPUT, axis=1, inplace=True)
    input_df.dropna()
    input_df[const.METABOLITES] = input_df[const.NAME_COLUMN] +'_'+ input_df[const.LABEL_COLUMN]
    input_df.drop([const.NAME_COLUMN,const.LABEL_COLUMN], axis=1, inplace=True)
    list_cols = input_df.columns.tolist()
    sample_col_first_list = list_cols[-1:] + list_cols[:-1]
    tabled_df =input_df.pivot(index= const.SAMPLE_COLUMN, columns= const.METABOLITES)[const.NA_CORRECTED_WITH_ZERO_COLUMN]
    tabled_df.reset_index(inplace=True)
    tabled_df = tabled_df.rename(index=str, columns = {const.SAMPLE_COLUMN:const.SAMPLE_ID})
    tabled_df = tabled_df.fillna(0)
    return tabled_df


def convert_quantfit_to_quickcheck_lcms(input_file_path):
    """
    This function converts Quantfit output to Quickcheck compatible input.

    :param input_file_path: (string) path of the file to be converted
    :return: (dataframe) Quickcheck compatible df
    """
    input_df = dat_hlp.read_file(input_file_path)
    input_df.drop(const.EXTRA_COLUMNS_LIST_QUANT_FIT_OUTPUT, axis=1, inplace=True)
    input_df.dropna()
    list_cols = input_df.columns.tolist()
    tabled_df =input_df.pivot(index=const.SAMPLE_NAME, columns=const.POINT_NAME)[const.INTENSITY_COLUMN]
    tabled_df.reset_index(inplace=True)
    tabled_df = tabled_df.rename(index=str, columns = {const.SAMPLE_NAME:const.SAMPLE_ID})
    tabled_df = tabled_df.fillna(0)
    return tabled_df


def convert_quantfit_to_metscape_lcms(input_file_path):
    """
    This function converts Quantfit output to Metscape compatible input.

    :param input_file_path: (string) path of the file to be converted
    :return: (dataframe) Metscape compatible df
    """
    input_df = dat_hlp.read_file(input_file_path)
    input_df.drop(const.EXTRA_COLUMNS_LIST_QUANT_FIT_OUTPUT_FOR_METSCAPE, axis=1, inplace=True)
    input_df.dropna()
    path = os.path.join(dir_path, const.CHEMICAL_FORMULA_FILE)
    metadata_df = pd.read_csv(path)
    metadata_df = metadata_df[[const.HMDB_ID, const.NAME_COLUMN]]
    metadata_df = metadata_df[metadata_df[const.HMDB_ID].notnull()][[const.HMDB_ID, const.NAME_COLUMN]]
    new_df = pd.merge(input_df, metadata_df, left_on= const.METABOLITE_NAME, right_on= const.NAME_COLUMN, how='inner')
    new_df.drop(const.NAME_COLUMN, axis=1, inplace=True)
    tabled_df =new_df.pivot(index= const.SAMPLE_NAME, columns=const.HMDB_ID)[const.INTENSITY_COLUMN]
    tabled_df.reset_index(inplace=True)
    tabled_df = tabled_df.rename(index=str, columns = {const.SAMPLE_NAME:const.SAMPLE_ID})
    tabled_df = tabled_df.fillna(0)
    return tabled_df

